import unittest

import scipy.stats as ss
from casex import *

from seedpod_ground_risk.path_analysis.descent_models.ballistic_model import BallisticModel
from seedpod_ground_risk.path_analysis.utils import bearing_to_angle


class BallisticModelPAEFTestCase(unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.ac = AircraftSpecs(enums.AircraftType.FIXED_WING,
                                2,  # width
                                1.8,  # length
                                7  # mass
                                )
        self.ac.set_ballistic_frontal_area(2)
        self.ac.set_glide_speed_ratio(15, 12)
        self.ac.set_glide_drag_coefficient(0.3)
        self.ac.set_ballistic_drag_coefficient(1.1)

        self.bm = BallisticDescent2ndOrderDragApproximation()
        self.bm.set_aircraft(self.ac)

    def test_ballistic_dist(self):
        """
        Test ballistic model impact distance distributions in the Path Aligned Event frame
        """
        alt_mean = 50
        alt_std = 5
        vx_mean = 18
        vx_std = 2.5
        vy = 1
        samples = 2000

        # Compute ballistic distances in the path aligned LTP frame with origin directly below the event location
        # AKA PAEF
        d_i, v_i, a_i, t_i = self.bm.compute_ballistic_distance(ss.norm(alt_mean, alt_std).rvs(samples),
                                                                ss.norm(vx_mean, vx_std).rvs(samples), vy)

        di_mean, di_std = ss.norm.fit(d_i)
        self.assertAlmostEqual(di_mean, 13.4, delta=0.5)
        self.assertAlmostEqual(di_std, 1.5, delta=0.2)

        import matplotlib.pyplot as mpl
        x = np.linspace(d_i.min(), d_i.max())
        y = ss.norm(di_mean, di_std).pdf(x)
        fig, ax = mpl.subplots(1, 1)
        ax.hist(d_i, density=True)
        ax.plot(x, y, 'r')
        ax.set_title('Event to Impact 1D Distance')
        ax.set_ylabel('Probability Density')
        ax.set_xlabel('Distance to Impact [m]')
        fig.show()

    def test_ballistic_time(self):
        """
        Test ballistic model impact time distributions in the Path Aligned Event frame.
        """
        alt_mean = 50
        alt_std = 5
        vx_mean = 18
        vx_std = 2.5
        vy = 1
        samples = 2000

        # Compute ballistic distances in the path aligned LTP frame with origin at the directly below the event location
        # AKA PAEF
        d_i, v_i, a_i, t_i = self.bm.compute_ballistic_distance(ss.norm(alt_mean, alt_std).rvs(samples),
                                                                ss.norm(vx_mean, vx_std).rvs(samples), vy)

        ti_mean, ti_std = ss.norm.fit(t_i)
        self.assertAlmostEqual(ti_mean, 7.4, delta=0.5)
        self.assertAlmostEqual(ti_std, 0.7, delta=0.2)

        import matplotlib.pyplot as mpl
        x = np.linspace(t_i.min(), t_i.max())
        y = ss.norm(ti_mean, ti_std).pdf(x)
        fig, ax = mpl.subplots(1, 1)
        ax.hist(t_i, density=True)
        ax.plot(x, y, 'r')
        ax.set_title('Event to Impact Time')
        ax.set_ylabel('Probability Density')
        ax.set_xlabel('Time to Impact [sec]')
        fig.show()

    def test_ballistic_vel(self):
        """
        Test ballistic model impact velocity distributions in the Path Aligned Event frame
        """
        alt_mean = 50
        alt_std = 5
        vx_mean = 18
        vx_std = 2.5
        vy = 1
        samples = 2000

        # Compute ballistic distances in the path aligned LTP frame with origin at the directly below the event location
        # AKA PAEF
        d_i, v_i, a_i, t_i = self.bm.compute_ballistic_distance(ss.norm(alt_mean, alt_std).rvs(samples),
                                                                ss.norm(vx_mean, vx_std).rvs(samples), vy)

        vi_mean, vi_std = ss.norm.fit(v_i)
        self.assertAlmostEqual(vi_mean, 7.1, delta=0.05)
        self.assertAlmostEqual(vi_std, 0.2, delta=0.2)

        import matplotlib.pyplot as mpl
        x = np.linspace(v_i.min(), v_i.max())
        y = ss.norm(vi_mean, vi_std).pdf(x)
        fig, ax = mpl.subplots(1, 1)
        ax.hist(v_i, density=True)
        ax.plot(x, y, 'r')
        ax.set_title('Impact Velocity')
        ax.set_ylabel('Probability Density')
        ax.set_xlabel('Impact Velocity [m/s]')
        fig.show()

    def test_ballistic_angle(self):
        """
        Test ballistic model impact angle distributions in the Path Aligned Event frame
        """
        alt_mean = 50
        alt_std = 5
        vx_mean = 18
        vx_std = 2.5
        vy = 1
        samples = 2000

        # Compute ballistic distances in the path aligned LTP frame with origin at the directly below the event location
        # AKA PAEF
        d_i, v_i, a_i, t_i = self.bm.compute_ballistic_distance(ss.norm(alt_mean, alt_std).rvs(samples),
                                                                ss.norm(vx_mean, vx_std).rvs(samples), vy)

        ai_mean, ai_std = ss.norm.fit(a_i)
        self.assertAlmostEqual(ai_mean, 1.57, delta=0.2)
        self.assertAlmostEqual(ai_std, 0.001, delta=0.002)

        import matplotlib.pyplot as mpl
        x = np.linspace(a_i.min(), a_i.max())
        y = ss.norm(ai_mean, ai_std).pdf(x)
        fig, ax = mpl.subplots(1, 1)
        ax.hist(a_i, density=True)
        ax.plot(x, y, 'r')
        ax.set_title('Impact Angle')
        ax.set_ylabel('Probability Density')
        ax.set_xlabel('Impact Angle [rad]')
        fig.show()


class BallisticModelNEDWindTestCase(unittest.TestCase):

    def setUp(self) -> None:
        super().setUp()
        self.ac = AircraftSpecs(enums.AircraftType.FIXED_WING,
                                0.88,  # width
                                0.88,  # length
                                3.75  # mass
                                )
        self.ac.set_ballistic_frontal_area(0.1)
        self.ac.set_glide_speed_ratio(16, 12)
        self.ac.set_glide_drag_coefficient(0.1)
        self.ac.set_ballistic_drag_coefficient(0.9)

        self.bm = BallisticDescent2ndOrderDragApproximation()
        self.bm.set_aircraft(self.ac)

    def test_ballistic_dist(self):
        """
        Profile ballistic model impact distance distributions in the North East Down frame with wind compensation
        """
        make_plot = True  # Set flag to plot result
        samples = 3000

        loc_x, loc_y = 0, 0

        # Conjure up our distributions for various things
        alt_mean = 50
        alt_std = 5

        vx_mean = 18
        vx_std = 2.5

        # In degrees!
        track_mean = 60
        track_std = 2

        # In degrees!
        wind_dir_mean = 120
        wind_dir_std = 5
        wind_vel_mean = 10
        wind_vel_std = 2

        alt = ss.norm(alt_mean, alt_std).rvs(samples)
        vel = ss.norm(vx_mean, vx_std).rvs(samples)
        track = np.deg2rad(ss.norm(track_mean, track_std).rvs(samples))
        wind_vel = ss.norm(wind_vel_mean, wind_vel_std).rvs(samples)
        wind_dir = bearing_to_angle(np.deg2rad(ss.norm(wind_dir_mean, wind_dir_std).rvs(samples)))

        wind_vel_x = wind_vel * np.cos(wind_dir)
        wind_vel_y = wind_vel * np.sin(wind_dir)

        bm = BallisticModel(self.ac)
        (means, cov), v_i, a_i = bm.transform(alt, vel, track, wind_vel_y, wind_vel_x, loc_x, loc_y)
        dist = ss.multivariate_normal(mean=means, cov=cov)

        if make_plot:
            # Make a sampling grid for plotting
            x, y = np.mgrid[(loc_x - 5):(loc_x + 95):0.5, (loc_y - 35):(loc_y + 40):0.5]
            pos = np.vstack([x.ravel(), y.ravel()])
            # Sample KDE PDF on these points
            density = dist.pdf(pos.T)
            # Plot sampled KDE PDF
            import matplotlib.pyplot as mpl
            fig, ax = mpl.subplots(1, 1, figsize=(8, 6))
            sc = ax.scatter(x, y, c=density)
            cbar = fig.colorbar(sc)
            cbar.set_label('Probability')
            ax.set_xlabel('Distance [m]')
            ax.set_ylabel('Distance [m]')
            ax.set_title(f'Ballistic Ground Impact Probability Density \n'
                         f' Altitude $\sim \mathcal{{N}}({alt_mean},{alt_std}^2)$m,'
                         f' Groundspeed $\sim \mathcal{{N}}({vx_mean},{vx_std}^2)$m/s,'
                         f' Track $\sim \mathcal{{N}}({track_mean},{track_std}^2)$deg,\n'
                         f' Wind speed $\sim \mathcal{{N}}({wind_vel_mean},{wind_vel_std}^2)$m/s,'
                         f' Wind bearing $\sim \mathcal{{N}}({wind_dir_mean},{wind_dir_std}^2)$deg')
            ax.arrow(loc_x, loc_y, vx_mean * np.cos(bearing_to_angle(np.deg2rad(track_mean))),
                     vx_mean * np.sin(bearing_to_angle(np.deg2rad(track_mean))), label='UAS Track', width=1,
                     color='blue')
            ax.arrow(loc_x, loc_y, wind_vel_mean * np.cos(bearing_to_angle(np.deg2rad(wind_dir_mean))),
                     wind_vel_mean * np.sin(bearing_to_angle(np.deg2rad(wind_dir_mean))), label='Wind Direction',
                     width=1, color='red')
            fig.show()


if __name__ == '__main__':
    unittest.main()
